@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 # Copyright 2009-2017 Yelp and Contributors
 # Copyright 2018-2019 Yelp
-# Copyright 2020 Affirm, Inc.
+# Copyright 2020 Affirm, Inc. and Contributors
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -710,7 +710,7 @@ class MRJobBinRunner(MRJobRunner):
 
         # save return code, turn off echo
         lines.append('# if script fails, print input URI before exiting')
-        lines.append('{ RETURNCODE=$?; set +x; } &> /dev/null')
+        lines.append('{ RETURNCODE=$?; set +x; } 1>&2 2>/dev/null')
         lines.append('')
 
         lines.append('{')
